mod error;
use symphonia::core::{
    audio::{Layout, SampleBuffer, SignalSpec},
    codecs::{Decoder, DecoderOptions, CODEC_TYPE_NULL},
    errors::Error as SymError,
    formats::{FormatReader, FormatOptions},
    io::MediaSourceStream,
    meta::MetadataOptions,
    probe::Hint
};
use tool_helper::{Error, Input};

pub type I16Samples = Vec::<i16>;
pub struct Settings {
    frequency: u32, 
    layout:    Layout
}

impl Settings {
    pub fn new(frequency: u32, layout: Layout) -> Settings {
        Settings{frequency, layout}
    }
}

pub fn load_as_i16_audio(input: Input, settings: Settings) -> Result<I16Samples, Error> {
    let media_stream = MediaSourceStream::new(Box::new(load_to_ram(input)?), Default::default());
    let format       = symphonia::default::get_probe().format(&Hint::new(), media_stream, &FormatOptions::default(), &MetadataOptions::default()).map_err(error::probe)?.format;
    let track        = format.tracks().iter().find(|t| t.codec_params.codec != CODEC_TYPE_NULL).ok_or_else(error::find_track)?;
    
    // Create a decoder for the track.
    let decoder  = symphonia::default::get_codecs().make(&track.codec_params, &DecoderOptions::default()).map_err(error::decoder)?;
    let track_id = track.id;

    decode(format, decoder, settings, track_id)
}

fn decode(mut format: Box<dyn FormatReader>, mut decoder: Box<dyn Decoder>, settings: Settings, track_id: u32) -> Result<I16Samples, Error> {
    let mut i16_audio  = Vec::new();
    let mut cur_sample = None;

    loop {
        // Get the next packet from the media format.
        let packet = match format.next_packet() {
            Ok(packet) => packet,
            Err(err) => {
                if let SymError::IoError(io_err) = &err {
                    if io_err.kind() == std::io::ErrorKind::UnexpectedEof {
                        return Ok(i16_audio);
                    }                    
                }
                return Err(error::next_packet(err));
            }
        };

        // Consume any new metadata that has been read since the last packet.
        format.metadata().skip_to_latest();

        // If the packet does not belong to the selected track, skip over it.
        if packet.track_id() != track_id {
            continue;
        }

        // Decode the packet into audio samples.
        let packet = decoder.decode(&packet).map_err(error::decode)?;
        if cur_sample.is_none() {
            let duration = packet.capacity() as u64;
            
            cur_sample = Some(SampleBuffer::<i16>::new(duration, SignalSpec::new_with_layout(settings.frequency, settings.layout)));
        }

        if let Some(cur_sample) = &mut cur_sample {
            cur_sample.copy_interleaved_ref(packet);

            for sample in cur_sample.samples() {
                i16_audio.push(*sample);
            }
        }        
    }
}

fn load_to_ram(mut input: Input) -> Result<std::io::Cursor<Vec<u8>>, Error> {
    let mut buffer = Vec::default();
    
    input.read_to_end(&mut buffer)?;
    Ok(std::io::Cursor::new(buffer))
}